"""Person information module"""
__all__ = ('Person',)

import re
from datetime import datetime

from .fields import Birthplace

_URL_ID_RE = re.compile(r'(?:(?:$|/)person/)?(\d+)-([^-]+)-([^/]+)')


class Person:
    """Person information. Need await object to fetch all else information will extract from url only."""
    __slots__ = ('image', 'name', 'name_transcription', 'id', 'career', 'birthday', 'birthplace', 'height', 'url')

    def __init__(self, url: str, *, name: str | None = None):
        """need await"""
        self.url = url
        self.height: float | None = None
        self.birthplace: Birthplace | None = None
        if m := _URL_ID_RE.search(url):
            id_, first, last = m.groups()
            self.id: int | None = int(id_)
            self.name_transcription = f'{first.capitalize()} {last.capitalize()}'
        else:
            self.name_transcription = ''
            self.id = None
        self.name = self.name_transcription if name is None else name

    def __await__(self):
        from .._utils import poster_and_soup
        if not (ps := (yield from poster_and_soup(self.url).__await__())):
            return self
        soup, self.image = ps
        if name := soup.select_one('[itemprop="name"]'):
            self.name = name.text
        if name := soup.select_one('[itemprop="alternativeHeadline"]'):
            self.name_transcription = name.text
        table = soup.select_one('.b-post__info')
        self.career = *(i.text for i in table.select('[itemprop="jobTitle"]')),
        date = table.select_one('[itemprop="birthDate"]').attrs.get('datetime')
        self.birthday = datetime.strptime(date, '%Y-%m-%d').date()
        sel = table.select('td.l+td:not(:has(*))')
        if not sel:
            return self
        birthplace, *height = sel
        self.height = float(height[0].text.strip().removesuffix('м')) if height else None
        match [i.strip() for i in birthplace.text.split(',')]:
            case [city, state, subcountry, country]:
                self.birthplace = Birthplace(country=country, city=city, subcountry=subcountry, state=state)
            case [city, state, country]:
                self.birthplace = Birthplace(country=country, city=city, state=state)
            case [city, country]:
                self.birthplace = Birthplace(country=country, city=city)
            case [country]:
                self.birthplace = Birthplace(country=country)
        return self

    def __repr__(self):
        return f"Person({f'{self.url!r}, ' if self.url else ''}{f'name={self.name!r}' if self.name else ''})"



"""Franchises information parsing module"""
__all__ = ('FranchiseInfo', 'FranchiseEntry')

from typing import NamedTuple

from bs4 import Tag, BeautifulSoup

from .fields import Hyperlink
from .._utils import hyperlink, get_any_int, poster_and_soup


class FranchiseEntry(NamedTuple):
    """ Franchise entry information """
    title: Hyperlink
    year: int
    rating: float | None = None


class FranchiseInfo:
    """ Franchise information class """
    __slots__ = ('url', '_entries', '_keys', '_current', '_poster')

    def __init__(self, url: str | None = None, soup: BeautifulSoup | Tag | None = None):
        """await needed"""
        self._entries: dict[int, FranchiseEntry] = {}
        self._keys: tuple[int, ...] = ()
        self._poster: str | None = None
        self._current: int | None = None
        self.url = url
        if soup is not None:
            self._setup_from_soup(soup)

    def __await__(self):
        if not (ps := (yield from poster_and_soup(self.url).__await__())):
            return self
        soup, self._poster = ps
        if not self.entries:
            self._setup_from_soup(soup)
        return self

    @property
    def entries(self) -> dict[int, FranchiseEntry]:
        """franchise entries (without current)"""
        return self._entries

    @entries.setter
    def entries(self, entries: dict[int, FranchiseEntry]):
        self._entries = entries
        self._keys = *entries.keys(),

    @property
    def poster(self):
        """wide poster, need await self first"""
        return self._poster

    @property
    def current(self):
        """Current position"""
        return self._current

    @property
    def previous(self):
        """Previous position"""
        return None if self._current == 1 else self.entries[self._current - 1]

    @property
    def next(self):
        """Next position"""
        return None if self._current == len(self.entries) - 1 else self.entries[self._current + 1]

    def _setup_from_soup(self, soup: BeautifulSoup | Tag):
        content = soup.select('.b-post__partcontent_item')[::-1]
        entries = {}
        for entry in content:
            n = int(entry.find(class_='num').text.strip())
            if 'current' in entry.attrs.get('class', ()):
                self._current = n
                continue
            entries[n] = FranchiseEntry(title=hyperlink(entry.select_one('.title>a')),
                                        year=get_any_int(entry.find(class_='year')),
                                        rating=None if (r := entry.find(class_='rating').text) == '—' else float(r))
        self.entries = entries

    def __len__(self):
        return len(self.entries)

    def __getitem__(self, item: int | slice) -> FranchiseEntry | tuple[FranchiseEntry] | tuple:
        if isinstance(item, int):
            return self.entries[self._keys[item]]
        if isinstance(item, slice):
            return *(self.entries[k] for k in self._keys[item]),
        raise TypeError(f'Expected int or slice; got {type(item).__name__} ({item!r})')

    def __str__(self):
        return f"{self.__class__.__qualname__}({repr(self.url) if self.url else ''})"



"""Module for parsing a Posts"""
__all__ = ('get_post_info',)

from bs4 import BeautifulSoup

from . import *
from ..info import PostInfo
from ..info.fields import *
from ..info.person import Person
from ..urls import Quality


def get_post_info(soup: BeautifulSoup, *, url: str | None = None) -> PostInfo:
    """Parse post info table from page soup"""
    if url is None:
        url = attr(soup.find('meta', property='og:video'), 'content', default='')
    title = text(soup.find('h1', itemprop='name'))
    tables = {i.find('h2').next_element.text: i.select_one('td:not(:has(h2))') for i in soup.select('tr:has(td>h2)')}
    get = lambda _k: tables.get(_k, empty_tag)
    ratings = {}
    for span in get('Рейтинги').select('span'):
        if a := span.find('a'):
            service = a.text
            url = parse_help_href(a['href'])
        else:
            service = span.next_element.text.strip().removesuffix(':')
            url = ''
        match span.attrs.get('class'):
            case ['b-post__info_rates', k, *_]:
                pass
            case [k, 'b-post__info_rates', *_]:
                pass
            case _:
                k = service.casefold()
        rating = float(text(span.find('span'), 0.))
        votes = get_any_int(text(span.find('i'), 0))
        ratings[k] = Rating(service=Hyperlink(name=service, url=url), rating=rating, votes=votes)
    rating_b = soup.find(class_='b-post__rating')
    ratings['hdrezka'] = Rating(service=Hyperlink(name='HDRezka', url=url),
                                rating=float(rating_b.find(itemprop='average').text),
                                votes=get_any_int(rating_b.find(itemprop='votes').text))
    release_tag = get('Дата выхода')
    return PostInfo(
        title=title,
        orig_title=text(soup.find(itemprop='alternativeHeadline'), title),
        poster=sidecover(soup),
        description=text(soup.find(class_='b-post__description_text')),
        duration=int(attr(soup.find('meta', property='og:duration'), 'content', default=0)),
        updated_time=int(attr(soup.find('meta', property='og:updated_time'), 'content', default=0)),
        ratings=ratings,
        rankings=(*(Rank(hyperlink(a), get_any_int(a.next_element.next_element))
                    for a in get('Входит в списки').select('a')),),
        slogan=text(get('Слоган')).removeprefix('«').removesuffix('»'),
        release=Release(day=text(release_tag.next_element),
                        year=int(text(release_tag.find('a'), '0').split(' ', 1)[0])),
        country=hyperlinks(get('Страна')),
        directors=(*(Person(a.attrs.get('href', ''), name=a.text) for a in get('Режиссер').select('a')),),
        genre=hyperlinks(get('Жанр')),
        quality=Quality(t) if (t := text(get('В качестве'))) else None,
        translators=(*map(str.strip, text(get('В переводе')).split(',')),),
        age_rating=(AgeRating(0) if (span := get('Возраст').find('span')) is None else
                    AgeRating(get_any_int(span.text), text(span.next_element.next_element))),
        duration_m=int(text(get('Время')).split(' ', 1)[0]),
        collections=hyperlinks(get('Из серии')),
        persons=(*(Person(a.attrs.get('href', ''), name=a.text) for a in soup.select('[itemprop="actor"]>a')),),
    )



"""API response types"""
__all__ = ('APIResponse', 'TrailerResponse')

from typing import TypedDict


class APIResponse(TypedDict):
    """Structure of the response to any request to the API"""
    success: bool
    message: str
    url: str
    quality: str
    subtitle: str
    subtitle_lns: dict[str, str]
    subtitle_def: str
    thumbnails: str


class TrailerResponse(TypedDict):
    """Structure of the response to trailer iframe request"""
    success: bool
    message: str
    code: str
    title: str
    description: str
    link: str



"""General user-friendly interface to HDRezka"""
__all__ = ('Player', 'PlayerBase', 'PlayerMovie', 'PlayerSeries')

from collections import defaultdict
from typing import Any, SupportsInt, Optional

from bs4 import BeautifulSoup

from ._cache import CACHE
from .._bs4 import BUILDER
from ..api.ajax import AJAX
from ..errors import UnknownContentType
from ..post import Post, urls_from_ajax_response, URLs


class PlayerBase:
    """Base type of Player"""
    __slots__ = ('post',)

    def __init__(self, url_or_cast: Any):
        """Need await"""
        if isinstance(url_or_cast, PlayerBase):
            self.post: Post = url_or_cast.post
            return
        elif not isinstance(url_or_cast, str):
            url_or_cast = str(url_or_cast)
        self.post = Post(url_or_cast)

    def __await__(self):
        """
        Initialize `self.post`
        Do not call twice!
        """
        yield from self.post.__await__()
        return self

    async def get_trailer_iframe(self) -> str:
        """Get trailer <iframe> HTML"""
        return (await AJAX.get_trailer_video(self.post.id)).get('code', '')

    def _translator(self, translator_id: Optional[SupportsInt] = None) -> int:
        if translator_id is None:
            return self.post.translator_id
        translator_id = int(translator_id)
        return self.post.translators.ids[abs(translator_id)] if translator_id <= 0 else translator_id

    def __repr__(self):
        return f'{self.__class__.__qualname__}({self.post.url!r})'


class PlayerMovie(PlayerBase):
    """Movies Player Type"""
    __slots__ = ()

    async def get_stream(self, translator_id: Optional[SupportsInt] = None) -> URLs:
        """Returns movie stream `URLs`"""
        return urls_from_ajax_response(await AJAX.get_movie(self.post.id, self._translator(translator_id)))


class PlayerSeries(PlayerBase):
    __slots__ = ()

    async def get_episodes(self, translator_id: Optional[SupportsInt] = None) -> defaultdict[int, tuple[int, ...]]:
        """Returns available episodes"""
        episodes = BeautifulSoup((await AJAX.get_episodes(self.post.id, self._translator(translator_id)))['episodes'],
                                 builder=BUILDER)
        result: defaultdict[int, tuple[int, ...]] = defaultdict(tuple)
        for i in episodes.find_all(class_='b-simple_episode__item', attrs=('data-season_id', 'data-episode_id')):
            result[int(i.attrs.get('data-season_id'))] += int(i.attrs.get('data-episode_id', '0')),
        return result

    async def get_stream(self, season: int, episode: int, translator_id: Optional[SupportsInt] = None) -> URLs:
        """Returns episode stream `URLs`"""
        return urls_from_ajax_response(
            await AJAX.get_stream(self.post.id, self._translator(translator_id), season, episode))


async def player(url_or_path: Any) -> PlayerMovie | PlayerSeries:
    """
    Returns either Player Series if series, or PlayerMovie if movie, otherwise raises UnknownContentType
    """
    cast = PlayerBase(url_or_path)
    cached = CACHE.get(cast.post.url)
    if cached is not None:
        return cached
    await cast
    value: PlayerMovie | PlayerSeries
    match cast.post.type:
        case 'tv_series':
            value = PlayerSeries(cast)
        case 'movie':
            value = PlayerMovie(cast)
        case _ as e:
            raise UnknownContentType(e)
    CACHE.set(cast.post.url, value)
    return value


Player = player  # as if the name of the class



"""Primitive cache storage"""
__all__ = ('CACHE',)


class _CacheStorage:
    __slots__ = ('max_size', '__storage', 'get')

    def __init__(self, max_size: int = 10):
        self.max_size = max_size
        self.__storage: dict = {}
        self.get = self.__storage.get

    def set(self, key, value):
        """Store some cache"""
        self.__storage[key] = value
        for k in (*self.__storage.keys(),):
            if len(self.__storage) <= self.max_size:
                break
            del self.__storage[k]


CACHE = _CacheStorage()



"""General post module"""
__all__ = ('Post',)

import urllib.parse

from bs4 import BeautifulSoup

from ._utils.parser import get_post_info
from .info.franchises import FranchiseInfo
from .._bs4 import BUILDER
from ..api.http import get_response
from ..translators import Translators
from ..url import Request


class Post:
    """Stores information about the post"""
    __slots__ = ('url', 'translator_id', 'id', 'name', 'type', 'info', 'translators', 'franchises')

    def __init__(self, url: str):
        """Need await"""
        if not urllib.parse.urlparse(url).hostname:
            url = Request.host_join(url)
        self.url = url

    def __await__(self):
        """
        Async initialize, prepare attributes.
        Do not call twice!
        """
        response = yield from get_response('GET', self.url).__await__()
        soup = BeautifulSoup(response.text, builder=BUILDER)
        self.type = soup.find('meta', property='og:type')['content'].removeprefix('video.')
        self.translator_id = self._get_translator_id(soup)
        self.info = get_post_info(soup, url=self.url)
        self.translators = self._get_translators(soup)

        self.id = int(soup.find(id='post_id')['value'])
        self.name = soup.find(class_='b-post__title').text.strip()
        franchises_url = soup.find(class_='b-post__franchise_link_title')
        self.franchises = FranchiseInfo(url=franchises_url and franchises_url.attrs.get('href'), soup=soup)
        return self

    def _get_translator_id(self, soup: BeautifulSoup) -> int | None:
        """self.type must exist"""
        init_cdn_obj = 'sof.tv.%s' % {'tv_series': 'initCDNSeriesEvents', 'movie': 'initCDNMoviesEvents'}[self.type]
        for script in soup.find_all(lambda tag: tag.name == 'script' and not tag.attrs and tag.string):
            if not (s := script.string):
                continue
            obj_i = s.find(init_cdn_obj)
            if obj_i != -1:
                s = s[obj_i + len(init_cdn_obj):].split(',', 2)[1].strip()
                if s.isnumeric():
                    return int(s)
        return None

    def _get_translators(self, soup: BeautifulSoup) -> Translators:
        arr = {i.text.strip(): int(i['data-translator_id']) for i in el.find_all(recursive=False) if i.text
               } if (el := soup.find(id='translators-list')) else {}
        if not arr:
            arr[self.info.translators[0]] = self.translator_id
        return Translators(arr)

    def __repr__(self):
        return f'{self.__class__.__qualname__}<{self.name!r}; {self.type!r}>'



r"""
# HDRezka site API

# Install

`pip install HDRezka`

# Example

```python
import asyncio
import os

from hdrezka import Search
from hdrezka.api.http import login_global


async def main():
    await login_global(os.environ['LOGIN_NAME'], os.environ['LOGIN_PASSWORD'])
    player = await (await Search('Breaking Bad').get_page(1))[0].player
    # or use url (await Player('series/thriller/646-vo-vse-tyazhkie-2008.html'))
    print(player.post.info, end='\n\n')

    translator_id = None  # default
    for name, id_ in player.post.translators.name_id.items():
        if 'субтитры' in name.casefold(): translator_id = id_; break

    stream = await player.get_stream(1, 1, translator_id)  # raise AJAXFail if invalid episode or translator
    video = stream.video
    print(await video.last_url[-1])  # best quality (.m3u8) (last source)
    # source - is not a finished link, await will return the link after the redirect
    print(await video[video.min].last_url[0].mp4, end='\n\n')  # worst quality (.mp4) (first source)

    subtitles = stream.subtitles
    print(subtitles.default.url)  # subtitles.ru.url or subtitles['Русский'].url


if __name__ == '__main__': asyncio.run(main())
```

# Bypass Blocking

## Custom HOST

```python
from hdrezka.url import Request

Request.HOST = 'https://hdrezka.club/'
```

> If the specified domain is protected by Cloudflare, successful response retrieval from the server cannot be guaranteed
> and may be inconsistent. See [this file](https://github.com/NIKDISSV-Forever/HDRezka/blob/main/mirrors.txt).

# Proxies

> pip install httpx[socks]
>
> tor

```python
import httpx

import hdrezka.api.http

hdrezka.api.http.DEFAULT_CLIENT = httpx.AsyncClient(
    headers={'User-Agent': 'Mozilla/5.0'},
    follow_redirects=True,
    proxy='socks5://localhost:9050'
)
```

## Log In (recommended for now)

Create an account on HDRezka.

If registration is temporarily disabled, try logging in via social media and resetting your password (so that you have
one at all).

Call `hdrezka.api.http.login_global` - this function will send a request to `hdrezka.api.http.Request.REDIRECT_URL`,
which will redirect to an active mirror. The function will then store the retrieved mirror globally in `Request.HOST`
and save the cookies in `hdrezka.api.http.DEFAULT_CLIENT`.

```python
import os

from hdrezka.api.http import login_global


async def main():
    await login_global(os.environ['LOGIN_NAME'], os.environ['LOGIN_PASSWORD'])
```

This method will also help if HDRezka considers your IP suspicious, suggests changing your VPN (proxy),
and returns a 403 error.
"""
__author__ = 'nikdissv'

from .api import *
from .post import *
from .stream import *



"""Basic wrapper for all http requests of the package"""
__all__ = ('get_response', 'login_global', 'DEFAULT_CLIENT', 'DEFAULT_REQUEST_KWARGS')

from typing import Any, Optional, TypedDict

import httpx

from ..url import Request


class RequestKwargs(TypedDict):
    """**kwargs for httpx.request"""
    content: Optional[Any]
    data: Optional[Any]
    files: Optional[Any]
    json: Optional[Any]
    params: Optional[Any]
    headers: Optional[Any]
    cookies: Optional[Any]
    auth: Optional[Any]
    follow_redirects: Optional[bool | Any]
    timeout: Optional[Any]
    extensions: Optional[Any]


DEFAULT_CLIENT = httpx.AsyncClient(headers={'User-Agent': 'Mozilla/5.0'})
DEFAULT_REQUEST_KWARGS: RequestKwargs | dict = {}


async def get_response(method: str, url: str | httpx.URL, **kwargs) -> httpx.Response:
    """
    passed **kwargs have more weight than DEFAULT_REQUEST_KWARGS.
    Returns DEFAULT_CLIENT.request(method, url, **kwargs, **DEFAULT_REQUEST_KWARGS)
    """
    for k, v in DEFAULT_REQUEST_KWARGS.items():
        if k not in kwargs:
            kwargs[k] = v
    return await DEFAULT_CLIENT.request(method, url, **kwargs)


async def login_global(email: str, password: str):
    """
    It enters the site using `DEFAULT_CLIENT`.
    Updates the global `HOST` to bypass the blocking, and `DEFAULT_CLIENT` cookies.
    """
    _follow_redirects = DEFAULT_CLIENT.follow_redirects
    DEFAULT_CLIENT.follow_redirects = False
    resp = await get_response('GET', Request.REDIRECT_URL)
    DEFAULT_CLIENT.follow_redirects = _follow_redirects
    url = httpx.URL(resp.headers.get('Location', resp.url))
    ajax_login_url = url.join('/ajax/login/')

    await get_response(
        'POST', ajax_login_url,
        data={'login_name': email, 'login_password': password, 'login_not_save': '0', 'login': 'submit'})
    Request.HOST = f'https://{url.host}'
